local _,N = ...

local form = "|T%s:0|t %s"
local line = form .. "\r"
local missing = "|cFFFF0000%s|r\r"
local have = "|cFF00FF00%s|r\r"
local gold = "|cFEFED000%s|r"


local function goldify(text)
  if not text then return "" end
  return text:gsub("([^\n\r\.?! ][^\n\r\.?!]+[\.?!]*)([\n\r]*)","|cFEFED000%1|r%2")
end

local function markItem(name)
  if name and #name > 0 then return name else return "--" end
end
N.markItem = markItem

local function findCurrency(match)
  for i=1,GetCurrencyListSize() do
    local name, isHeader, _, _, _, count = GetCurrencyListInfo(i)
    if not isHeader and match == name then
      return count
    end
  end
  return 0
end

----------------------------------------------
-- Adapted from Blizzard's QuestFrame.lua
----------------------------------------------

local function color(str, enough)
  return format(enough and have or missing, str)
end
local function count(num, texture, str, enough)
  return "- " .. (num > 1 and num .. " " or "") .. format(form, texture, str)
end

local function objectives(id)
  if not id then return end
  local tasks = {}
  local i = 1
  while true do
    local text, objectiveType, finished, have, need = GetQuestObjectiveInfo(id, i, false)
    if not text or objectiveType == "log" then return tasks end
      tinsert(tasks, { text=text, type = objectiveType, finished = finished, have = have, need = need })
    i = i + 1
  end
end

local function matchObjectives(showFinished, tasks, name, texture, s, append, complete)
  if tasks and name and texture then
    for _,info in pairs(tasks) do
      if not info.matched then
        local text = info.text
        local ind = text:find(name)
        if ind then
          info.matched = true
          info.text = text:sub(0, ind-1) .. "|T" .. texture .. ":0|t" .. text:sub(ind)
          if not showFinished and info.finished then
            return s .. color(append, true)
          else
            return s
          end
        end
      end
    end
  end
  return s .. color(append, complete)
end

local function required()
  local showFinished = N.Handler.core.finished
  local tasks = objectives(GetQuestID())
  local numRequiredItems = GetNumQuestItems()
  local numRequiredCurrencies = GetNumQuestCurrencies()
  local requiredMoney = GetQuestMoneyToGet()
  if (not tasks or not next(tasks)) and (numRequiredItems or 0) + (numRequiredCurrencies or 0) + (requiredMoney or 0) == 0 then return "" end
  local s = ""
  if requiredMoney > 0 then
    s = s .. color("- " .. GetMoneyString(requiredMoney), requiredMoney <= GetMoney())
  end
  for i=1, numRequiredItems do
    if (IsQuestItemHidden(i) == 0) then
      local name, texture, numItems = GetQuestItemInfo("required", i)
      s = matchObjectives(showFinished, tasks, name, texture, s, count(numItems, texture, name), GetItemCount(name) >= numItems)
    end
  end
  for i=1, numRequiredCurrencies do
    local name, texture, numItems = GetQuestCurrencyInfo("required", i)
    s = matchObjectives(showFinished, tasks, name, texture, s, count(numItems, texture, name), findCurrency(name) >= numItems)
  end

  local taskText = ""
  for _,info in pairs(tasks or {}) do
    if showFinished or not info.finished then
      taskText = taskText .. color("- " .. info.text, info.finished)
    end
  end

  return "\r" .. s ..  taskText
end
----------------------------------------------
-- Adapted from Blizzard's QuestInfo.lua
----------------------------------------------
local function AddSpellToBucket(spellBuckets, type, rewardSpellIndex)
  if not spellBuckets[type] then
    spellBuckets[type] = {}
  end

  tinsert(spellBuckets[type], rewardSpellIndex)
end

local function getRewards(prefix)
  local f = line
  local pref = prefix and "+ " or ""
  local numQuestRewards = GetNumQuestRewards()
  local numQuestChoices = GetNumQuestChoices()
  local numQuestCurrencies = GetNumRewardCurrencies()
  local money = GetRewardMoney()
  local skillName, skillIcon, skillPoints = GetRewardSkillPoints()
  local xp = GetRewardXP()
  local artifactXP, artifactCategory = GetRewardArtifactXP()
  local honor = GetRewardHonor()
  local playerTitle = GetRewardTitle()
  local numSpellRewards = GetNumRewardSpells()
  local numQuestSpellRewards = 0
  local spellGetter = GetRewardSpell

  for rewardSpellIndex = 1, numSpellRewards do
    local texture, name, isTradeskillSpell, isSpellLearned, hideSpellLearnText, isBoostSpell, garrFollowerID, spellID = spellGetter(rewardSpellIndex)
    local knownSpell = IsSpellKnownOrOverridesKnown(spellID)

    -- only allow the spell reward if user can learn it
    if texture and not knownSpell and (not isBoostSpell or IsCharacterNewlyBoosted()) and (not garrFollowerID or not C_Garrison.IsFollowerCollected(garrFollowerID)) then
      numQuestSpellRewards = numQuestSpellRewards + 1
    end
  end

  local totalRewards = numQuestRewards + numQuestChoices + numQuestCurrencies
  if totalRewards == 0 and money == 0 and xp == 0 and not playerTitle and numQuestSpellRewards == 0 and artifactXP == 0 then
    return nil
  end

  local name, texture, quality, isUsable, numItems
  local rewardsCount = 0

  local s = ""

  if ( numQuestSpellRewards > 0 ) then
    local spellBuckets = {}

    for rewardSpellIndex = 1, numSpellRewards do
      local texture, name, isTradeskillSpell, isSpellLearned, hideSpellLearnText, isBoostSpell, garrFollowerID, spellID = spellGetter(rewardSpellIndex)
      local knownSpell = IsSpellKnownOrOverridesKnown(spellID)
      if texture and not knownSpell and (not isBoostSpell or IsCharacterNewlyBoosted()) and (not garrFollowerID or not C_Garrison.IsFollowerCollected(garrFollowerID)) then
        if ( isTradeskillSpell ) then
          AddSpellToBucket(spellBuckets, QUEST_SPELL_REWARD_TYPE_TRADESKILL_SPELL, rewardSpellIndex)
        elseif ( isBoostSpell ) then
          AddSpellToBucket(spellBuckets, QUEST_SPELL_REWARD_TYPE_ABILITY, rewardSpellIndex)
        elseif ( garrFollowerID ) then
          AddSpellToBucket(spellBuckets, QUEST_SPELL_REWARD_TYPE_FOLLOWER, rewardSpellIndex)
        elseif ( not isSpellLearned ) then
          AddSpellToBucket(spellBuckets, QUEST_SPELL_REWARD_TYPE_AURA, rewardSpellIndex)
        else
          AddSpellToBucket(spellBuckets, QUEST_SPELL_REWARD_TYPE_SPELL, rewardSpellIndex)
        end
      end
    end

    for orderIndex, spellBucketType in ipairs(QUEST_INFO_SPELL_REWARD_ORDERING) do
      local spellBucket = spellBuckets[spellBucketType]
      if spellBucket then
        for i, rewardSpellIndex in ipairs(spellBucket) do
          local texture, name, isTradeskillSpell, isSpellLearned, _, isBoostSpell, garrFollowerID, spellID = spellGetter(rewardSpellIndex)
          if i == 1 then
            s = s .. format(gold, QUEST_INFO_SPELL_REWARD_TO_HEADER[spellBucketType]) .. "\r"
          end

          if garrFollowerID then
            local followerInfo = C_Garrison.GetFollowerInfo(garrFollowerID)
            s = s .. pref .. format(f, followerInfo.portraitIconID, "Follower: " .. followerInfo.name)
          else
            s = s .. pref .. format(f, texture, GetSpellLink(spellID) or name)
          end
        end
        s = s .. "\n"
      end
    end
  end

  -- Item rewards
  local index
  local baseIndex = rewardsCount
  local buttonIndex = 0
  for i = 1, numQuestRewards, 1 do
    buttonIndex = buttonIndex + 1
    index = i + baseIndex
    name, texture, numItems, quality, isUsable = GetQuestItemInfo("reward", i)
    name = markItem(name)
    local color = isUsable and select(4, GetItemQualityColor(quality)) or "ffff0000"
    local displayName = GetQuestItemLink("reward", i) or name
    s = s .. pref
    if numItems > 1 then s = s .. numItems .. " " end
    displayName = "|c" .. color .. displayName .. "|r"
    s = s .. format(f, texture, displayName)
  end
  if numQuestRewards > 0 then s = s .. "\r" end
  -- currency
  baseIndex = rewardsCount
  local foundCurrencies = 0
  buttonIndex = buttonIndex + 1
  for i = 1, GetMaxRewardCurrencies(), 1 do
    index = i + baseIndex
    name, texture, numItems = GetQuestCurrencyInfo("reward", i)
    if name and texture and numItems then
      s = s .. pref
      if numItems > 1 then s = s .. numItems .. " " end
      s = s .. format(f, texture, name)
      rewardsCount = rewardsCount + 1
      foundCurrencies = foundCurrencies + 1
      buttonIndex = buttonIndex + 1
      if foundCurrencies == numQuestCurrencies then
        break
      end
    end
  end
  if numQuestRewards + foundCurrencies > 1 then s = s .. "\n" end

  if ( honor > 0 ) then
    local icon
    if (UnitFactionGroup("player") == PLAYER_FACTION_GROUP[0]) then
      icon = "Interface\\Icons\\PVPCurrency-Honor-Horde"
    else
      icon = "Interface\\Icons\\PVPCurrency-Honor-Alliance"
    end
    s = s .. pref .. format(f, icon, HONOR .. ": " .. BreakUpLargeNumbers(honor)) .. "\r\r"
  end

  -- Title reward
  if playerTitle then
    s = s .. pref .. format(gold, REWARD_TITLE) .. playerTitle .. "\r\r"
  end

  if artifactXP > 0 then
    local name, icon = C_ArtifactUI.GetArtifactXPRewardTargetInfo(artifactCategory)
    s = s .. pref .. format(f, icon, BreakUpLargeNumbers(artifactXP)) .. "\r\r"
  end

  if money > 0 then
    s = s .. pref .. GetMoneyString(money) .. "\r\r"
  end
  -- XP rewards
  if xp and xp > 0 then
    s = s .. pref .. format(f,"Interface\\ICONS\\xp_icon",BreakUpLargeNumbers(xp)) .. "\r"
  end
  -- Skill Point rewards
  if skillPoints then
    s = s .. pref .. format(f, skillIcon, skillPoints .. " " .. skillName) .. "\r"
  end

  if numQuestChoices > 0 then
    if numQuestChoices > 1 then
      s = s .. "\n" .. format(gold, REWARD_CHOICES) .. "\r"
    end
    local index
    local baseIndex = rewardsCount
    for i = 1, numQuestChoices do
      name, texture, numItems, quality, isUsable = GetQuestItemInfo("choice", i)
      name = markItem(name)
      local color = isUsable and select(4, GetItemQualityColor(quality)) or "ffff0000"
      local displayName = GetQuestItemLink("choice", i) or name
      s = s .. pref
      if numItems > 1 then s = s .. numItems .. " " end
      displayName = "|c" .. color .. displayName .. "|r"
      s = s .. format(f, texture, displayName)
      rewardsCount = rewardsCount + 1
    end
    s = s .. "\n"
  end

  return s
end
N.getRewards = getRewards

local function plusObjective(event)
  local text
  if event == "QUEST_PROGRESS" then
    local id = GetQuestID()
    if id then
      text = select(2, GetQuestLogQuestText(GetQuestLogIndexByID(id)))
    end
    text = goldify(text) .. required()
  else
    text = goldify(GetObjectiveText())
  end
  if N.Handler.core.rewards then
    text = text .. "\r" .. getRewards(true)
  end
  if not text or #text == 0 then return "" end
  return "\n" .. text
end
N.plusObjective = plusObjective