local _,N = ...
local lastEvent
local speed = 0.5

local plusObjective = N.plusObjective

local ANIM_FREEZE = 3
local ANIM_SAY = 60
local ANIM_YELL = 64
local ANIM_ASK = 65
local ANIM_READ = 520

local function AnimFinished(self)
  if self.anim == ANIM_FREEZE or self.anim == ANIM_READ then return end
  if self.shouldLoop then
    self:SetAnimation(self.anim, 0)
  else
    self.anim = nil
  end
end

local function SetupAnimations(self, anim)
  if self.player or self.guid == UnitGUID("player") then
    anim = ANIM_READ
  elseif self.freeze then
    anim = ANIM_FREEZE
  end
  if self.anim == anim then return end
  self.anim = anim
  local should = N.Handler.core.animate
  if anim == nil or self.freeze or should == "never" then
    self.shouldLoop = false
    self:StopAnimating()
    return
  end

  self:SetAnimation(anim, 0)
  self.shouldLoop = should == "loop"
end

local function queueItems()
  for i=1,GetNumQuestRewards() do
    GetQuestItemInfo("reward", i)
  end
  for i=1,GetNumQuestChoices() do
    GetQuestItemInfo("choice", i)
  end
end

local function ShowLink(self,_,link)
  GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT", 10, 0)
  GameTooltip:SetHyperlink(link)
  GameTooltip:Show()
end

local frame = Conversation

function frame:AnchorAlerts(conf)
  local shouldAnchor = conf.adjustAlerts
  if shouldAnchor then
    if self.alertAnchor then return end
    self:ClearAllPoints()
    self.alertAnchor = AlertFrame:AddJustAnchorFrameSubSystem(self.TopAnchor)
    local alertAnchor = self.alertAnchor
    alertAnchor.AdjustAnchors = function(_, relativeAlert)
      if not self:IsShown() then return relativeAlert end
      if relativeAlert == AlertFrame then
        self:SetPoint("BOTTOM", 0, 96)
      else
        self:SetPoint("BOTTOM", relativeAlert, "TOP", 0, 10)
      end
      return self.TopAnchor
    end
    AlertFrame:SetSubSustemAnchorPriority(alertAnchor, 1)
  else
    local anchor = self.alertAnchor
    if not anchor then return end
    self.alertAnchor = nil
    local systems = AlertFrame.alertFrameSubSystems
    for i,v in ipairs(systems) do
      if v == anchor then
        tremove(systems, i)
        return
      end
    end
    N.updateFrame(self, conf)
  end
end

frame:SetScript("OnLoad", function(self)
  self:RegisterEvent("LOADING_SCREEN_ENABLED")
  self:RegisterEvent("GOSSIP_SHOW")
  self:RegisterEvent("GOSSIP_CLOSED")
  self:RegisterEvent("QUEST_GREETING")
  self:RegisterEvent("QUEST_FINISHED")
  self:RegisterEvent("QUEST_DETAIL")
  self:RegisterEvent("QUEST_PROGRESS")
  self:RegisterEvent("QUEST_COMPLETE")
  self:RegisterEvent("QUEST_ITEM_UPDATE")
  --self:RegisterForClicks("LeftButtonUp")
  self:EnableKeyboard(true)

  self.NameFrame.Name:SetPoint("TOPLEFT", self.PortraitFrame.Portrait, "TOPRIGHT", 2, -19)

  local text = self.TextFrame
  text:SetScript("OnHyperlinkEnter", ShowLink)
  text:EnableMouse(true)
  text:SetScript("OnMouseDown", function(_, button)
    if button == "LeftButton" then
      self:Next()
      return true
    end
    return false
  end)

  for i=1,select("#", self:GetChildren()) do
    local child = select(i, self:GetChildren())
    for j=1,select("#", child:GetAnimationGroups()) do
      local animGroup = select(j, child:GetAnimationGroups())
      for k=1,select("#", animGroup:GetAnimations()) do
        local anim = select(k, animGroup:GetAnimations())
        anim:SetStartDelay(anim:GetStartDelay() * speed)
        anim:SetDuration(anim:GetDuration() * speed)
      end
    end
  end

  local model = self.MainFrame.Model
  model.SetupAnimations = SetupAnimations
  model:SetPortraitZoom(0.85)

  model:SetScript("OnAnimFinished", AnimFinished)

  self.TextFrame.Progress:SetScript("OnFinished", function()
    self:Next()
  end)
end)

function frame:SetTop(anchor)
  self.TopAnchor:SetPoint("TOP", anchor or self)
end

local delay

local function Delay(funcName, ...)
  local wait
  if GetGossipOptions() or GetGossipActiveQuests() or GetGossipAvailableQuests() then
    wait = 0.1
  else wait = 0.3 end
  local root = ...
  local n = select("#", ...)
  local t = { ... }
  delay = funcName
  C_Timer.After(wait, function()
    if delay ~= funcName then return end
    root[funcName](unpack(t,1,n))
  end)
end

frame:SetScript("OnEvent", function(self, event, ...)
  N.Handler:Output(event)
  if event == "LOADING_SCREEN_ENABLED" then
    self:Reset()
    self:CloseImmediately()
    return
  elseif event == "GOSSIP_SHOW" then
    if GetNumGossipAvailableQuests() == 0 and GetNumGossipActiveQuests() == 0 and GetNumGossipOptions() == 1 then
      if select(2, GetGossipOptions()) ~= "gossip" then
        SelectGossipOption(1)
        return
      end
    end
    Delay("PlayCurrent", self, GetUnitName("npc"), GetGossipText())
  elseif event == "GOSSIP_CLOSED" then
    Delay("Close", self, true)
  elseif event == "QUEST_GREETING" then
    Delay("PlayCurrent", self, GetUnitName("questnpc") or GetUnitName("npc"), GetGreetingText())
  elseif event == "QUEST_FINISHED" then
    Delay("Close", self)
  elseif event == "QUEST_DETAIL" then
    if QuestIsFromAdventureMap() or QuestIsFromAreaTrigger() or (... ~= nil and ... ~= 0) then
      return
    end
    queueItems()
    Delay("PlayCurrent", self, GetTitleText(), GetQuestText() .. plusObjective(event))
  elseif event == "QUEST_PROGRESS" then
    queueItems()
    Delay("PlayCurrent", self, GetTitleText(), GetProgressText() .. plusObjective(event))
  elseif event == "QUEST_COMPLETE" then
    Delay("PlayCurrent", self, GetTitleText(), GetRewardText())
  elseif event == "QUEST_ITEM_UPDATE" then
    return
  end
  lastEvent = event
end)
local concealed
frame:SetScript("OnShow", function(self)
  if N.Handler.core.conceal then
    UIParent:Hide()
    concealed = true
  end
  self.TopAnchor:Show()
  AlertFrame:UpdateAnchors()
end)
frame:SetScript("OnHide", function(self)
  if concealed then
    concealed = nil
    UIParent:Show()
  end
  self.TopAnchor:Hide()
  AlertFrame:UpdateAnchors()
end)

local function endInteraction(partial)
  N.Handler:Output("Closing ",partial)
  if not partial then
    CloseGossip()
    CloseQuestChoice()
  end
  CloseQuest()
end

frame:SetScript("OnKeyDown", function(self, key)
  local conf = N.Handler.core
  if key == conf.next then
    self:SetPropagateKeyboardInput(false)
    self:Next()
    return
  elseif key == conf.back then
    self:SetPropagateKeyboardInput(false)
    self:FadeText(self.full)
    return
  elseif key == conf.close then
    endInteraction()
  elseif key == conf.auto then
    local child = self.ChoicesFrame.auto
    if child then
      child:GetScript("OnEnter")(child)
      child:Click()
    else
      self:Next()
    end
    self:SetPropagateKeyboardInput(false)
    return
  else
    local keyNumber
    for i=1,12 do
      if conf[tostring(i)] == key then
        keyNumber = i
        break
      end
    end
    local sum = self.ChoicesFrame.sum
    if keyNumber and keyNumber <= sum then
      local child = select(sum + 1 - keyNumber, self.ChoicesFrame:GetChildren())
      if child.choose then
        child:GetScript("OnEnter")(child)
        child:Click()
      end
      self:SetPropagateKeyboardInput(false)
      return
    end
  end

  self:SetPropagateKeyboardInput(true)
  return
end)

function frame:CloseImmediately()
  if (self.finishTimer) then
    self.finishTimer:Cancel()
    self.finishTimer = nil
  end
  self:Hide()
  endInteraction()
end

function frame:FadeinFrames()
  self.MainFrame.TalkingHeadsInAnim:Play()
  C_Timer.After(0.5 * speed, function()
    self.NameFrame.Fadein:Play()
  end)
  C_Timer.After(0.75 * speed, function()
    self.TextFrame.Fadein:Play()
  end)
  self.BackgroundFrame.Fadein:Play()
  self.PortraitFrame.Fadein:Play()
end

function frame:FadeoutFrames()
  if concealed then
    concealed = nil
    UIParent:Show()
  end
  self.ChoicesFrame:Close()
  self.MainFrame.Close:Play()
  self.NameFrame.Close:Play()
  self.TextFrame.Close:Play()
  self.BackgroundFrame.Close:Play()
  self.PortraitFrame.Close:Play()
  if self.TextFrame.Continue:GetAlpha() ~= 0 then
    self.TextFrame.FadeoutContinue:Play()
  end
  self.TextFrame.Progress:Stop()
end

function frame:SetText(text)
  if not text or #text == 0 then
    self.TextFrame.Text:SetText()
    return
  end
  text = text:trim()
  local line = text:find("\n")
  if line then
    self.remaining = text:sub(line)
    text = text:sub(1,line):trim()
  else
    self.remaining = ""
  end
  local t = self.TextFrame.Text
  t:SetText(text)
  if text:sub(1,1) ~= "|" then
    local anim
    if text:find("?") then
      anim = ANIM_ASK
    elseif text:find("!") then
      anim = ANIM_YELL
    else
      anim = ANIM_SAY
    end
    self.MainFrame.Model:SetupAnimations(anim)
  end
  local trail
  for _,delim in ipairs{ "\n\r", "\.?!", ",:"} do
    local sub = t:GetText()
    while t:IsTruncated() do
      sub, trail = select(3, sub:find(
        "(.+[" .. delim .. "])(.+[^\s" .. delim .. "].+)"
      ))
      if not sub then break end
      sub = sub:trim()
      if #sub == 0 then break end
      self.remaining = trail .. self.remaining
      t:SetText(sub)
    end
  end
  self.remaining = self.remaining:trim()
  if not self.remaining or #self.remaining == 0 then
    if self.TextFrame.Continue:GetAlpha() ~= 0 then
      self.TextFrame.FadeoutContinue:Play()
    end
    C_Timer.After(0.5 * speed, function()
      self.ChoicesFrame:Refresh(lastEvent)
    end)
  else
    if self.remaining:sub(1,2) == "|c" then
      self.TextFrame.Continue:SetText(self.remaining:sub(1,10) .. "...|r")
    else
      self.TextFrame.Continue:SetText("...")
    end
    if self.TextFrame.Continue:GetAlpha() == 0 then
      self.TextFrame.FadeinContinue:Play()
    end
    if N.Handler.core.slideshow then
      self.TextFrame.Progress:Restart()
    end
    self.ChoicesFrame:Close()
  end
end

function frame:FadeText(text)
  self.TextFrame.Fadeout:Play()
  C_Timer.After(0.25 * speed, function()
    self:SetText(text)
  end)
  C_Timer.After(0.5 * speed, function()
    self.TextFrame.Fadein:Play()
  end)
end

function frame:Reset(text, name)
  self:StopAnimating()
  self.BackgroundFrame.TextBackground:SetAlpha(0.01)
  self.NameFrame.Name:SetAlpha(0.01)
  self.TextFrame.Text:SetAlpha(0.01)
  self.MainFrame.Sheen:SetAlpha(0.01)
  self.MainFrame.TextSheen:SetAlpha(0.01)

  self.MainFrame.Model:SetAlpha(0.01)
  self.MainFrame.Model.PortraitBg:SetAlpha(0.01)
  self.PortraitFrame.Portrait:SetAlpha(0.01)
  self.MainFrame.Overlay.Glow_LeftBar:SetAlpha(0.01)
  self.MainFrame.Overlay.Glow_RightBar:SetAlpha(0.01)
  self.MainFrame.CloseButton:SetAlpha(0.01)

  self.MainFrame:SetAlpha(1)
  self.NameFrame.Name:SetText(name)
  self:SetText(text)
  --self.TextFrame.Text:SetText(text)
end

function frame:PlayCurrent(name, text, new)
  if not new and (not self:IsShown() or self.MainFrame.CloseButton:GetAlpha() ~= 1) then new = true end
  local model = self.MainFrame.Model

  if( self.finishTimer ) then
    self.finishTimer:Cancel()
    self.finishTimer = nil
  end

  local textFormatted = string.format(text)
  self.full = textFormatted
  if ( text ) then
    self:Show()

    local unit = "questnpc"
    local guid = UnitGUID(unit)
    if not guid then
      unit = "npc"
      guid = UnitGUID(unit)
    end
    model.freeze = UnitIsDead(unit)

    if ( model.guid ~= guid ) then
      model.anim = nil
      if new == nil then new = true end
      model.guid = guid
      if guid:sub(1,4) == "Item" then
        model.player = true
        model:SetUnit("player")
      else
        model.player = false
        model:SetUnit(unit)
      end
      model:SetupAnimations()
    end

    if ( new ) then
      self.ChoicesFrame:Hide()
      self:Reset(textFormatted, name)
      self:FadeinFrames()
    else
      if ( name ~= self.NameFrame.Name:GetText() ) then
        -- Fade out the old name and fade in the new name
        self.NameFrame.Fadeout:Play()
        C_Timer.After(0.25 * speed, function()
          self.NameFrame.Name:SetText(name)
        end)
        C_Timer.After(0.5 * speed, function()
          self.NameFrame.Fadein:Play()
        end)

        --/self.MainFrame.TalkingHeadsInAnim:Play()
      end

      self:FadeText(textFormatted)
    end
  end
end

function frame:Next()
  if self.MainFrame.CloseButton:GetAlpha() ~= 1 then return end
  if self.remaining and #(self.remaining) > 0 then
    self:FadeText(self.remaining)
  elseif self.ChoicesFrame.sum == 0 then
    endInteraction()
  end
end

function frame:Close(partial)
  self.MainFrame.Model:StopAnimating()
  self:FadeoutFrames()
  self.finishTimer = nil
  endInteraction(partial)
end

frame:GetScript("OnLoad")(frame)